/***********************************************************************/
/*!

    AudioScience HPI driver
    Copyright (C) 1997-2010  AudioScience Inc. <support@audioscience.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of version 2 of the GNU General Public License as
    published by the Free Software Foundation;

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


\file
Functions for reading DSP code to load into DSP

(Linux only:) If DSPCODE_FIRMWARE_LOADER is defined, code is read using
hotplug firmware loader from individual dsp code files

If neither of the above is defined, code is read from linked arrays.
DSPCODE_ARRAY is defined.

HPI_INCLUDE_**** must be defined
and the appropriate hzz?????.c or hex?????.c linked in

 */
/***********************************************************************/
#define SOURCEFILE_NAME "hpidspcd.c"
#include "hpidspcd.h"
#include "hpidebug.h"

#ifndef HPI_BUILD_KERNEL_MODE
#include <string.h>
#include <stdlib.h>
#endif

/**
 Header structure for binary dsp code file (see asidsp.doc)
 This structure must match that used in s2bin.c for generation of asidsp.bin
 */

#ifndef DISABLE_PRAGMA_PACK1
#pragma pack(push, 1)
#endif

struct code_header{
	uint32_t size;
	char type[4];
	uint32_t adapter;
	uint32_t version;
	uint32_t crc;
};

#ifndef DISABLE_PRAGMA_PACK1
#pragma pack(pop)
#endif

#ifdef HPI_BUILD_DSPCODE_FIRMWARE
#define HPI_VER_DECIMAL ((int)(HPI_VER_MAJOR(HPI_VER) * 10000 + \
            HPI_VER_MINOR(HPI_VER) * 100 + HPI_VER_RELEASE(HPI_VER)))

/***********************************************************************/
#include "linux/pci.h"
/*-------------------------------------------------------------------*/
short HpiDspCode_Open(uint32_t nAdapter,
		      struct dsp_code *psDspCode, uint32_t *pdwOsErrorCode)
{
	const struct firmware *psFirmware = psDspCode->psFirmware;
	struct code_header header;
	char fw_name[20];
	int err;

	sprintf(fw_name, "asihpi/dsp%04x.bin", nAdapter);

	err = request_firmware(&psFirmware, fw_name, &psDspCode->psDev->dev);

	if (err != 0) {
		dev_printk(KERN_ERR,  &psDspCode->psDev->dev,
		"%d, request_firmware failed for  %s\n",
			       err, fw_name);
		goto error1;
	}
	if (psFirmware->size < sizeof(header)) {
		dev_printk(KERN_ERR,  &psDspCode->psDev->dev,
		"Header size too small %s\n", fw_name);
		goto error2;
	}
	memcpy(&header, psFirmware->data, sizeof(header));
	if (header.adapter != nAdapter) {
		dev_printk(KERN_ERR,  &psDspCode->psDev->dev,
		"Adapter type incorrect %4x != %4x\n",
			       header.adapter, nAdapter);
		goto error2;
	}
	if (header.size != psFirmware->size) {
		dev_printk(KERN_ERR,  &psDspCode->psDev->dev,
				"Code size wrong  %d != %ld\n",
			       header.size, (unsigned long)psFirmware->size);
		goto error2;
	}

	if (header.version / 100 != HPI_VER_DECIMAL / 100) {
		dev_printk(KERN_ERR, &psDspCode->psDev->dev,
		    "Incompatible firmware version "
		    "DSP image %d != Driver %d\n",
		    header.version, HPI_VER_DECIMAL);
		goto error2;
	}

	if (header.version != HPI_VER_DECIMAL) {
		dev_printk(KERN_WARNING, &psDspCode->psDev->dev,
		    "Firmware: release version mismatch  DSP image %d != Driver %d\n",
		    header.version, HPI_VER_DECIMAL);
	}

	HPI_DEBUG_LOG(DEBUG, "Dsp code %s opened\n", fw_name);
	psDspCode->psFirmware = psFirmware;
	psDspCode->dwBlockLength = header.size / sizeof(uint32_t);
	psDspCode->dwWordCount = sizeof(header) / sizeof(uint32_t);
	psDspCode->dwVersion = header.version;
	psDspCode->dwCrc = header.crc;
	return 0;

error2:
	release_firmware(psFirmware);
error1:
	psDspCode->psFirmware = NULL;
	psDspCode->dwBlockLength = 0;
	return HPI_ERROR_DSP_FILE_NOT_FOUND;
}

/*-------------------------------------------------------------------*/
void HpiDspCode_Close(struct dsp_code *psDspCode)
{
	if (psDspCode->psFirmware != NULL) {
		HPI_DEBUG_LOG(DEBUG, "Dsp code closed\n");
		release_firmware(psDspCode->psFirmware);
		psDspCode->psFirmware = NULL;
	}
}

/*-------------------------------------------------------------------*/
void HpiDspCode_Rewind(struct dsp_code *psDspCode)
{
	/* Go back to start of  data, after header */
	psDspCode->dwWordCount = sizeof(struct code_header) / sizeof(uint32_t);
}

/*-------------------------------------------------------------------*/
short HpiDspCode_ReadWord(struct dsp_code *psDspCode, uint32_t *pdwWord)
{
	if (psDspCode->dwWordCount + 1 > psDspCode->dwBlockLength)
		return HPI_ERROR_DSP_FILE_FORMAT;

	*pdwWord =
	    ((uint32_t *)(psDspCode->psFirmware->data))[psDspCode->dwWordCount];
	psDspCode->dwWordCount++;
	return 0;
}

/*-------------------------------------------------------------------*/
short HpiDspCode_ReadBlock(size_t nWordsRequested,
			   struct dsp_code *psDspCode, uint32_t **ppdwBlock)
{
	if (psDspCode->dwWordCount + nWordsRequested >
	    psDspCode->dwBlockLength)
		return HPI_ERROR_DSP_FILE_FORMAT;

	*ppdwBlock =
	    ((uint32_t *)(psDspCode->psFirmware->data)) + psDspCode->dwWordCount;
	psDspCode->dwWordCount += nWordsRequested;
	return 0;
}
#else				//not  defined HPI_OS_LINUX
/********************************************************************/
/* Load dspcode from file asidsp.bin */
#ifdef HPI_BUILD_DSPCODE_FILE
/*-------------------------------------------------------------------*/

short HpiDspCode_Open(uint32_t nAdapter,
		      struct dsp_code *psDspCode, uint32_t *pdwOsErrorCode)
{
	uint16_t wError;
	int i, nRead = 0;
	struct code_header header;
	const char *pszFilepath;

	pszFilepath = HpiOs_GetDspCodePath(nAdapter);

	wError =
	    HpiOs_fopen_rb(pszFilepath, &psDspCode->pDspCodeFile,
			   pdwOsErrorCode);
	if (wError != 0)
		return (wError);

	// seek for first CODE header
	for (i = 0; i < 1000; i++) {
		HpiOs_fseek(psDspCode->pDspCodeFile, i, SEEK_SET);
		nRead = (int)HpiOs_fread(&header, 1, sizeof(header),
				psDspCode->pDspCodeFile);
		if (nRead < sizeof(header)) {
			HPI_DEBUG_LOG0(DEBUG, "Ran out of data\n");
			break;
		}
		if (strncmp(header.type, "CODE", 4) == 0)
			break;
	}
	if ((nRead < sizeof(header)) || (i == 1000)) {
		HpiOs_fclose(psDspCode->pDspCodeFile);
		HPI_DEBUG_LOG0(DEBUG, "CODE header not found\n");
		return (HPI_ERROR_DSP_FILE_NO_HEADER);
	}
	// walk CODE headers searching for correct one
	while ((nRead == sizeof(header)) && (header.adapter != nAdapter)) {
		HPI_DEBUG_LOG1(VERBOSE, "Skipping %04lx ",
			       header.adapter);
		if (header.size == 0) {
			HpiOs_fclose(psDspCode->pDspCodeFile);
			HPI_DEBUG_LOG0(DEBUG,
				       "NULL header found\n");
			return (HPI_ERROR_DSP_FILE_NULL_HEADER);
		}
		HpiOs_fseek(psDspCode->pDspCodeFile,
			    (header.size - sizeof(header)), SEEK_CUR);
		nRead =(int)HpiOs_fread(&header, 1, sizeof(header),
				psDspCode->pDspCodeFile);
	}
	if ((nRead < sizeof(header)) || (header.adapter != nAdapter)) {
		HpiOs_fclose(psDspCode->pDspCodeFile);
		HPI_DEBUG_LOG1(ERROR,
			       "Adapter type %04x not found\n",
			       nAdapter);
		return (HPI_ERROR_DSP_SECTION_NOT_FOUND);
	}
	else
	{
		HPI_DEBUG_LOG1(VERBOSE,
			       "Adapter type %04x code found\n",
			       nAdapter);
	}

	psDspCode->aCodeBuffer =
	    (uint32_t *)HpiOs_MemAlloc(sizeof(uint32_t) * BLOCK_LIMIT_DWORD);
	if (!psDspCode->aCodeBuffer) {
		HpiOs_fclose(psDspCode->pDspCodeFile);
		HPI_DEBUG_LOG0(ERROR, "Out of memory error\n");
		return (HPI_ERROR_DSP_FILE_OTHER_ERROR);
	}
	psDspCode->nAdapter = nAdapter;
	psDspCode->dwBlockLength = (header.size - sizeof(header)) / 4;
	psDspCode->dwWordCount = 0;
	psDspCode->dwVersion = header.version;
	psDspCode->dwCrc = header.crc;
	return 0;
}

/*-------------------------------------------------------------------*/
void HpiDspCode_Rewind(struct dsp_code *psDspCode)
{
	HpiDspCode_Close(psDspCode);
	HpiDspCode_Open(psDspCode->nAdapter, psDspCode, NULL);
}

/*-------------------------------------------------------------------*/
void HpiDspCode_Close(struct dsp_code *psDspCode)
{
	HpiOs_MemFree(psDspCode->aCodeBuffer);
	psDspCode->aCodeBuffer = NULL;
	if (psDspCode->pDspCodeFile != NO_FILE) {
		HpiOs_fclose(psDspCode->pDspCodeFile);
		psDspCode->pDspCodeFile = NO_FILE;
	}
}

/*-------------------------------------------------------------------*/
short HpiDspCode_ReadWord(struct dsp_code *psDspCode, uint32_t *pdwWord)
{
	size_t count;
	count =
	    HpiOs_fread(pdwWord, sizeof(uint32_t), 1, psDspCode->pDspCodeFile);
	if (count != 1)
		return (HPI_ERROR_DSP_FILE_FORMAT + 21);	// 231
	// could also calculate checksum here?
	psDspCode->dwWordCount++;
	if (psDspCode->dwWordCount > psDspCode->dwBlockLength)
		return (HPI_ERROR_DSP_FILE_FORMAT + 22);	// 232
	return 0;
}

/*-------------------------------------------------------------------*/
short HpiDspCode_ReadBlock(size_t nWordsRequested,
			   struct dsp_code *psDspCode, uint32_t **ppdwBlock)
{
	size_t count;

	if (nWordsRequested > BLOCK_LIMIT_DWORD)
		return (HPI_ERROR_DSP_FILE_FORMAT + 23);	// 233

	count =
	    HpiOs_fread(psDspCode->aCodeBuffer, sizeof(uint32_t), nWordsRequested,
			psDspCode->pDspCodeFile);
	if (count != nWordsRequested)
		return (HPI_ERROR_DSP_FILE_FORMAT + 24);	// 234

	psDspCode->dwWordCount += (long)nWordsRequested;
	if (psDspCode->dwWordCount > psDspCode->dwBlockLength)
		return (HPI_ERROR_DSP_FILE_FORMAT + 25);	// 235

	*ppdwBlock = psDspCode->aCodeBuffer;
	return (0);
}
#endif				//defined HPI_DSPCODE_FILE
#endif				//not  defined HPI_OS_LINUX

