/***********************************************************************/
/**

    AudioScience HPI driver
    Copyright (C) 1997-2010  AudioScience Inc. <support@audioscience.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of version 2 of the GNU General Public License as
    published by the Free Software Foundation;

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

\file
Functions for reading DSP code to load into DSP

 hpi_dspcode_defines HPI DSP code loading method
Define exactly one of these to select how the DSP code is supplied to
the adapter.

End users writing applications that use the HPI interface do not have to
use any of the below defines; they are only necessary for building drivers

HPI_DSPCODE_FILE:
DSP code is supplied as a file that is opened and read from by the driver.

HPI_DSPCODE_FIRMWARE:
DSP code is read using the hotplug firmware loader module.
     Only valid when compiling the HPI kernel driver under Linux.
*/
/***********************************************************************/
#ifndef _HPIDSPCD_H_
#define _HPIDSPCD_H_

#include "hpi_internal.h"

#ifndef DISABLE_PRAGMA_PACK1
#pragma pack(push, 1)
#endif

#ifdef HPI_BUILD_DSPCODE_FIRMWARE
/** Descriptor for dspcode from firmware loader */
struct dsp_code {
	/**  Firmware descriptor */
	const struct firmware *psFirmware;
	struct pci_dev *psDev;
        /** Expected number of words in the whole dsp code,INCL header */
	long int dwBlockLength;
        /** Number of words read so far */
	long int dwWordCount;
        /** Version read from dsp code file */
	uint32_t dwVersion;
        /** CRC read from dsp code file */
	uint32_t dwCrc;
};

#elif defined HPI_BUILD_DSPCODE_FILE
	/* must match or be greater than the same constant in S2BIN.C */
#define BLOCK_LIMIT_DWORD 2048L

/** Descriptor used when dsp code read from a file */
struct dsp_code {
	HpiOs_FILE pDspCodeFile;	//!< File descriptor for dsp code file
	long int dwBlockLength;	//!< Expected number of words in the whole dsp code, from header
	long int dwWordCount;	//!< Number of words read so far
	uint32_t nAdapter;		//!< Adapter type
	uint32_t dwVersion;		//<! Version read from dsp code file
	uint32_t dwCrc;		//<! CRC read from dsp code file
	uint32_t *aCodeBuffer;
};

#else
#error Must define HPI_BUILD_DSPCODE_FIRMWARE, HPI_BUILD_DSPCODE_FILE in build environment
#endif

#ifndef DISABLE_PRAGMA_PACK1
#pragma pack(pop)
#endif

/** Prepare *psDspCode to refer to the requuested adapter.
 Searches the file, or selects the appropriate linked array

\return 0 for success, or error code if requested code is not available
*/
short HpiDspCode_Open(
	/// Code identifier, usually adapter family
	uint32_t nAdapter,
	/// Pointer to DSP code control structure
	struct dsp_code *psDspCode,
        /// Pointer to dword to receive OS specific error code
	uint32_t *pdwOsErrorCode
    );

/** Close the DSP code file */
void HpiDspCode_Close(struct dsp_code *psDspCode);

/** Rewind to the beginning of the DSP code file (for verify) */
void HpiDspCode_Rewind(struct dsp_code *psDspCode);

/** Read one word from the dsp code file
	\return 0 for success, or error code if eof, or block length exceeded
*/
short HpiDspCode_ReadWord(
	struct dsp_code *psDspCode, ///< DSP code descriptor
	uint32_t *pdwWord ///< Where to store the read word
);

/** Get a block of dsp code into an internal buffer, and provide a pointer to
that buffer. (If dsp code is already an array in memory, it is referenced,
not copied.)

\return Error if requested number of words are not available
*/
short HpiDspCode_ReadBlock(size_t nWordsRequested,
			   struct dsp_code *psDspCode,
      	// Pointer to store (Pointer to code buffer)
			   uint32_t **ppdwBlock
    );

#endif

