/***********************************************************************/
/*!
$Header: /Repository/drv/hpi/hpimux.c,v 1.47.2.1 2011/02/16 18:27:52 as-age Exp $

\file
Functions routing HPI messages to two or more modules (i.e. hpioct32 and hpiudp)

(c) AudioScience, Inc., 2007

 */
/***********************************************************************/

#ifdef HPI_BUILD_MULTIINTERFACE
#define SOURCEFILE_NAME "hpimux.c"

#include "hpi_internal.h"
#include "hpimsginit.h"
#include "hpidebug.h"

uint16_t HPI_DriverOpenIoctl(void);
void HPI_DriverCloseIoctl(void);
void HPI_MessageIoctl(HPI_MESSAGE *phm, HPI_RESPONSE *phr);

#if defined HPI_OS_LINUX && defined HPI_BUILD_SSX2
#include "hpissx2.h"
#define HPI_MESSAGE_LOCAL_ADAPTERS(m,r) HPI_MESSAGE_LOWER_LAYER(m,r,NULL)
#else
#define HPI_MESSAGE_LOCAL_ADAPTERS(m,r) HPI_MessageIoctl(m,r)
#endif
typedef uint16_t (*PHPI_DRIVEROPEN) (void);
typedef void (*PHPI_DRIVERCLOSE) (void);
typedef void (*PHPI_MESSAGE) (HPI_MESSAGE *phm, HPI_RESPONSE *phr,unsigned int timeout);

struct set_host_if {
	int pending;
	HPI_MESSAGE msg;
	char ip[256];
};

struct set_host_if host_if;

#ifdef HPI_BUILD_USEUDPDLL
#ifdef _WIN64
#define HPI_UDPDLLNAME	"asihpiudp64.dll"
#define HPI_DRIVEROPEN_SYMBOL "HPI_DriverOpenUdp"
#define HPI_DRIVERCLOSE_SYMBOL "HPI_DriverCloseUdp"
#define HPI_MESSAGEUDP_SYMBOL "HPI_MessageUdp"
#else
#define HPI_UDPDLLNAME	"asihpiudp32.dll"
#define HPI_DRIVEROPEN_SYMBOL "_HPI_DriverOpenUdp@0"
#define HPI_DRIVERCLOSE_SYMBOL "_HPI_DriverCloseUdp@0"
#define HPI_MESSAGEUDP_SYMBOL "_HPI_MessageUdp@12"
#endif
static uint16_t HPI_DriverOpenStub(void)
{
	return 0;
}

static void HPI_DriverCloseStub(void)
{
}

static void HPI_MessageStub(HPI_MESSAGE *phm, HPI_RESPONSE *phr,unsigned int timeout)
{
	HPI_InitResponse(phr, phm->wObject, phm->wFunction, HPI_ERROR_INVALID_OBJ_INDEX);
	if (phm->wFunction == HPI_SUBSYS_SET_NETWORK_INTERFACE) {
		host_if.pending = 1;
		memcpy( &host_if.msg, phm, sizeof(HPI_MESSAGE) );
		strncpy( host_if.ip, phm->u.s.resource.r.net_if, sizeof(host_if.ip) );
		host_if.msg.u.s.resource.r.net_if = host_if.ip;
		phr->wError = 0;
	}
}

static PHPI_DRIVEROPEN HPIMUX_DriverOpenUdp = HPI_DriverOpenStub;
static PHPI_DRIVERCLOSE HPIMUX_DriverCloseUdp = HPI_DriverCloseStub;
static PHPI_MESSAGE HPIMUX_MessageUdp = HPI_MessageStub;
static HMODULE hUdpModule = NULL;
static int udpLoadAlreadyFailed = 0;
#else
/* static linkage to hpiudp */
uint16_t HPI_DriverOpenUdp(void);
void HPI_DriverCloseUdp(void);
void HPI_MessageUdp(HPI_MESSAGE *phm, HPI_RESPONSE *phr, unsigned int timeout);

static PHPI_DRIVEROPEN HPIMUX_DriverOpenUdp = HPI_DriverOpenUdp;
static PHPI_DRIVERCLOSE HPIMUX_DriverCloseUdp = HPI_DriverCloseUdp;
static PHPI_MESSAGE HPIMUX_MessageUdp = HPI_MessageUdp;
#endif

static int nIoctlDriver = 0;
static int nUdpDriver = 0;
static int init = 0;

static uint16_t wNumIoctlAdapters = 0;
static uint16_t wNumUdpAdapters = 0;


uint16_t HPI_DriverOpen(void)
{
	HPI_DEBUG_LOG0(DEBUG,"HPI_DriverOpen()\n");
	nIoctlDriver = HPI_DriverOpenIoctl();
#if defined HPI_OS_LINUX
#if defined HPI_BUILD_SSX2
	if (nIoctlDriver) {
		struct hpi_message hm;
		struct hpi_response hr;
		int i;

		char * psSsxEnable;

		psSsxEnable = getenv("LIBHPI_SSX_ENABLE");
		if (psSsxEnable) {
			for (i = 0; i < strlen(psSsxEnable); i++) {
				if (psSsxEnable[i] == '1') {
					HPI_DEBUG_LOG1(INFO, "SSX2 enabled on adapter %d\n", i);
					HPI_InitMessageResponse( &hm, &hr,  HPI_OBJ_ADAPTER, HPI_ADAPTER_SET_PROPERTY );
					hm.u.ax.property_set.wProperty = HPI_ADAPTER_PROPERTY_ENABLE_SSX2;
					hm.u.ax.property_set.wParameter1 = 1;
					hm.u.ax.property_set.wParameter2 = 1;
					hm.wAdapterIndex = i;
					HPI_MESSAGE_LOCAL_ADAPTERS( &hm, &hr);
				}
			}
		}
	}
#endif
#endif

#ifdef HPI_BUILD_USEUDPDLL
	if (!udpLoadAlreadyFailed && !hUdpModule) {
		hUdpModule = LoadLibrary(TEXT(HPI_UDPDLLNAME));
		if (!hUdpModule) {
			HPI_DEBUG_LOG1(DEBUG,"Could not load %s\n",TEXT(HPI_UDPDLLNAME));
			HPI_DEBUG_LOG0(DEBUG, "UDP DLL Not Found\n");
			udpLoadAlreadyFailed = 1;
		} else {
			HPI_DEBUG_LOG1(DEBUG,
				"Successfully loaded %s\n",TEXT(HPI_UDPDLLNAME));
			HPI_DEBUG_LOG0(DEBUG, "UDP DLL Loaded\n");
			HPIMUX_DriverOpenUdp =
			    (PHPI_DRIVEROPEN) GetProcAddress(hUdpModule,
							     HPI_DRIVEROPEN_SYMBOL);
			if (HPIMUX_DriverOpenUdp)
				HPI_DEBUG_LOG0(DEBUG,
					"Successfully got proc address for _HPI_DriverOpenUdp@0\n");
			HPIMUX_DriverCloseUdp =
			    (PHPI_DRIVERCLOSE) GetProcAddress(hUdpModule,
							      HPI_DRIVERCLOSE_SYMBOL);
			HPIMUX_MessageUdp =
			    (PHPI_MESSAGE) GetProcAddress(hUdpModule,
							  HPI_MESSAGEUDP_SYMBOL);
		}
	}
#endif

	if (HPIMUX_DriverOpenUdp) {
		HPI_RESPONSE hr;
		nUdpDriver = HPIMUX_DriverOpenUdp();

		if( host_if.pending ) {
			HPIMUX_MessageUdp( &host_if.msg, &hr, HPI_ETHERNET_TIMEOUT_MS);
			host_if.pending = 0;
		}
	}

	HPI_DEBUG_LOG2(DEBUG,"HPI_DriverOpen ioctl=%d, udp=%d\n", nIoctlDriver, nUdpDriver);
	return (nIoctlDriver || nUdpDriver);
}

void HPI_DriverClose(void)
{
	HPI_DEBUG_LOG0(DEBUG,"HPI_DriverClose()\n");
	if (nIoctlDriver) {
		HPI_DEBUG_LOG0(DEBUG, "HPI_DriverCloseIoctl()\n");
		HPI_DriverCloseIoctl();
	}
	if (nUdpDriver) {
		HPI_DEBUG_LOG0(DEBUG, "HPI_DriverCloseUdp()\n");
		nUdpDriver = 0;
		HPIMUX_DriverCloseUdp();
#ifdef HPI_BUILD_USEUDPDLL
		FreeLibrary(hUdpModule);

		/* from load initialization */
		HPIMUX_DriverOpenUdp = HPI_DriverOpenStub;
		HPIMUX_DriverCloseUdp = HPI_DriverCloseStub;
		HPIMUX_MessageUdp = HPI_MessageStub;
		hUdpModule = NULL;
		udpLoadAlreadyFailed = 0;
#endif
	}
}

// This wrapper is required by hpifirmware.c  - TFE
void HPI_MessageUDP(HPI_MESSAGE *phm, HPI_RESPONSE *phr, const unsigned int nTimeout)
{
	HPIMUX_MessageUdp(phm, phr, nTimeout);
}

void GetNumAdapters(void)
{
	HPI_MESSAGE hm;
	HPI_RESPONSE hr_Ioctl;
	HPI_RESPONSE hr_Udp;

	HPI_InitMessageResponse(&hm, &hr_Ioctl, HPI_OBJ_SUBSYSTEM, HPI_SUBSYS_GET_NUM_ADAPTERS);
	HPI_InitResponse(&hr_Udp, HPI_OBJ_SUBSYSTEM, HPI_SUBSYS_GET_NUM_ADAPTERS, 0);

	if (nIoctlDriver) {
		HPI_DEBUG_LOG0(DEBUG,
			       "HPI_SUBSYS_GET_NUM_ADAPTERS - HPI_MESSAGE_LOCAL_ADAPTERS\n");
		HPI_MESSAGE_LOCAL_ADAPTERS(&hm, &hr_Ioctl);
		if (!hr_Ioctl.wError)
			wNumIoctlAdapters = hr_Ioctl.u.s.wNumAdapters;
		else
			wNumIoctlAdapters = 0;
	} else
		wNumIoctlAdapters = 0;

	if (nUdpDriver) {
		HPI_DEBUG_LOG0(DEBUG,
			       "HPI_SUBSYS_GET_NUM_ADAPTERS - HPI_MessageUdp\n");
		HPIMUX_MessageUdp(&hm, &hr_Udp, HPI_ETHERNET_TIMEOUT_MS);
		if (!hr_Udp.wError)
			wNumUdpAdapters = hr_Udp.u.s.wNumAdapters;
		else
			wNumUdpAdapters = 0;
	} else
		wNumUdpAdapters = 0;

	HPI_DEBUG_LOG2(DEBUG,
		       "HPI_SUBSYS_GET_NUM_ADAPTERS: Ioctl %d, Udp %d\n",
		       wNumIoctlAdapters,
		       wNumUdpAdapters);
}

void HPI_Message(HPI_MESSAGE *phm, HPI_RESPONSE *phr)
{
	HPI_DEBUG_MESSAGE(DEBUG, phm);

	if( !init ) {
		init = 1;
		memset(	&host_if, 0, sizeof(host_if) );
	}

	// check for various global messages
	if (phm->wObject == HPI_OBJ_SUBSYSTEM) {
		HPI_RESPONSE hr_Ioctl;
		HPI_RESPONSE hr_Udp;

		HPI_InitResponse(&hr_Ioctl, phm->wObject, phm->wFunction, 0);
		HPI_InitResponse(&hr_Udp, phm->wObject, phm->wFunction, 0);

		switch (phm->wFunction) {
			// subsys open - init the debug function
		case HPI_SUBSYS_OPEN: {
				HPI_InitResponse(phr, HPI_OBJ_SUBSYSTEM, HPI_SUBSYS_OPEN, 0);
				if (nIoctlDriver) {
					HPI_DEBUG_LOG0(DEBUG,
						"HPI_SUBSYS_OPEN - HPI_MESSAGE_LOCAL_ADAPTERS\n");
					HPI_MESSAGE_LOCAL_ADAPTERS(phm, &hr_Ioctl);
				}
				if (nUdpDriver) {
					HPI_DEBUG_LOG0(DEBUG,
						"HPI_SUBSYS_OPEN - HPI_MessageUdp\n");
					HPIMUX_MessageUdp(phm, &hr_Udp, HPI_ETHERNET_TIMEOUT_MS);
					if (hr_Udp.wError) {
						HPI_DEBUG_LOG0(DEBUG,
							"HPI_SUBSYS_OPEN - nUdpDriver = 0\n");
						nUdpDriver = 0;
					}
				}
			}
			break;

		case HPI_SUBSYS_CLOSE:
			HPI_InitResponse(phr, phm->wObject, phm->wFunction, 0);
			if (nIoctlDriver) {
				HPI_DEBUG_LOG0(DEBUG,
					       "HPI_SUBSYS_CLOSE - HPI_MESSAGE_LOCAL_ADAPTERS\n");
				HPI_MESSAGE_LOCAL_ADAPTERS(phm, phr);
			}
			if (nUdpDriver) {
				HPI_DEBUG_LOG0(DEBUG,
					       "HPI_SUBSYS_CLOSE - HPI_MessageUdp\n");
				HPIMUX_MessageUdp(phm, phr, HPI_ETHERNET_TIMEOUT_MS);
			}
			break;

		case HPI_SUBSYS_GET_VERSION:
			HPI_InitResponse(phr, phm->wObject, phm->wFunction, 0);
			phr->u.s.dwVersion = HPI_VER >> 8;	// break major.minor
			phr->u.s.dwData = HPI_VER;	// break major.minor.release
			break;

		case HPI_SUBSYS_GET_NUM_ADAPTERS:
			HPI_InitResponse(phr, phm->wObject, phm->wFunction, 0);
			GetNumAdapters();
			phr->u.s.wNumAdapters = wNumIoctlAdapters + wNumUdpAdapters;
			break;

		case HPI_SUBSYS_GET_ADAPTER:
			GetNumAdapters();
			if (phm->wObjIndex < wNumIoctlAdapters) {
				HPI_DEBUG_LOG0(DEBUG,
					       "HPI_SUBSYS_GET_ADAPTER - HPI_MESSAGE_LOCAL_ADAPTERS\n");
				HPI_MESSAGE_LOCAL_ADAPTERS(phm, phr);
			} else if (phm->wObjIndex - wNumIoctlAdapters <
				   wNumUdpAdapters) {
				phm->wObjIndex -= wNumIoctlAdapters;
				HPI_DEBUG_LOG0(DEBUG,
					       "HPI_SUBSYS_GET_ADAPTER - HPI_MessageUdp\n");
				HPIMUX_MessageUdp(phm, phr, HPI_ETHERNET_TIMEOUT_MS);
			} else
				HPI_InitResponse(phr, phm->wObject, phm->wFunction, HPI_ERROR_INVALID_OBJ_INDEX);
			break;

		case HPI_SUBSYS_SET_NETWORK_INTERFACE:
			HPIMUX_MessageUdp(phm, phr, HPI_ETHERNET_TIMEOUT_MS);
			break;

		case HPI_SUBSYS_OPTION_INFO:
		case HPI_SUBSYS_OPTION_GET:
		case HPI_SUBSYS_OPTION_SET:
			/* so far, only HPIUDP uses entity messages,
			but in future may require routing */
			HPIMUX_MessageUdp(phm, phr, HPI_ETHERNET_TIMEOUT_MS);
			break;

		default:
			/* Need to specifically determine which backend(s)
			receive subsys messages */
			HPI_DEBUG_LOG0(WARNING, "unrouted subsys message\n");
			HPI_InitResponse(phr, phm->wObject, phm->wFunction, HPI_ERROR_INVALID_FUNC);
			break;
		}
		return;
	}

	if (phm->wAdapterIndex >= 1000) {
		HPI_DEBUG_LOG0(DEBUG, "DEFAULT - HPI_MessageUdp\n");
		HPIMUX_MessageUdp(phm, phr, HPI_ETHERNET_TIMEOUT_MS);
	} else {
		HPI_DEBUG_LOG0(DEBUG, "DEFAULT - HPI_MESSAGE_LOCAL_ADAPTERS\n");
		HPI_MESSAGE_LOCAL_ADAPTERS(phm, phr);
	}
}
#endif
/*
$Log: hpimux.c,v $
Revision 1.47.2.1  2011/02/16 18:27:52  as-age
merge from HEAD

Revision 1.48  2011/02/16 18:24:36  as-age
correct HPI_SUBSYS_SET_NETWORK_INTERFACE error return - detected by using asihpitest.c in conjunction with ASI2416.

Revision 1.47  2011/01/06 16:52:28  as-age
Fix for a side effect related to issue#2728. Applications like asihpitest.c and hpi.py
assume they can call HPI_SubSysSetHotsNetworkInterface() followed by HPI_SubSysCreate()
and have network device be discovered. This is how 4.04 worked. This change
supports caching of SubSysSetNetworkInterface() so that the message is sent after
the driver is opened. There should be no effect on user applications.

Revision 1.46  2011/01/06 16:11:42  as-age
add dbg

Revision 1.45  2010/12/06 19:43:21  as-ewb
Fix non-dll build.

Revision 1.44  2010/12/02 21:37:59  as-age
belt and suspenders UDP DLL unload.

Revision 1.43  2010/12/02 21:23:19  as-age
fix udp unload (Eliot please review). Issue#3144

Revision 1.42  2010/11/23 02:36:16  as-ewb
Convert most conditional compilation define names to HPI_BUILD_x

Revision 1.41  2010/10/06 16:57:34  as-age
remove canUseUdpDriver, and yes, we tested first !

Revision 1.40  2010/10/05 21:53:42  as-ewb
fix cut/paste error

Revision 1.39  2010/10/05 21:08:47  as-ewb
Heed error returns from subsys get num adapters

Revision 1.38  2010/10/05 04:06:37  as-ewb
Remove hpi_network_ functions ids.
Change function ids subsys_entity_x to subsys_option_x

Revision 1.37  2010/10/04 23:13:27  as-ewb
Maybe fix issue#3099
On windows, must call SUBSYS_SET_NETWORK_INTERFACE
before udp adapters can be found.

Revision 1.36  2010/10/04 20:00:39  as-ewb
Factor out GetNumAdapters so it can be used by SubsysGetAdapter
also. (removes calling sequence dependency)

Revision 1.35  2010/09/28 15:46:46  as-tfe
Added udpLoadAlreadyFailed and HPI_MessageUDP().

Revision 1.34  2010/09/27 22:38:54  as-ewb
Load dll and call HPI_DriverOpenUdp in DriverOpen function.
(hpiudp doesn't start network discovery until it is required by
an hpi message)
Issue#2728

Revision 1.33  2010/09/22 23:27:29  as-ewb
add comment to subsys_get_adapter case

Revision 1.32  2010/09/20 22:34:45  as-ewb
Return error for subsys messages not explicitly routed.

Revision 1.31  2010/09/16 23:26:16  as-ewb
Don't overwrite phr when response buffer is being passed
down to a lower layer. Issue#2728

Revision 1.30  2010/09/09 03:04:32  as-ewb
hpi_find_adapters messages are no longer required

Revision 1.29  2009/12/10 21:00:27  as-tfe
Fix DLL and import names for 64 bit build.

Revision 1.28  2009/11/05 23:05:09  as-ewb
Correct function prototype. Issue#2428

Revision 1.27  2009/10/27 22:36:18  as-ewb
SSX2 support linux only

Revision 1.26  2009/10/27 21:56:03  as-ewb
Support linux hpimux + hpissx2 build. Issue#2162

Revision 1.25  2009/08/13 04:48:05  as-ewb
Allow non dll build.
Don't fail driver open even if local driver open fails (network may work)

Revision 1.24  2009/06/04 23:25:56  as-ewb
merge from branch_hpimsg1

Revision 1.23.2.2  2009/05/26 16:47:19  as-tfe
typo

Revision 1.23.2.1  2009/05/26 16:41:34  as-age
change how wSize is set

Revision 1.23  2009/04/17 14:52:33  as-tfe
HPI_SUBSYS_CLOSE sets nUdpDriver so ASI2416's don't continue to show
up in asictrl after switching to local adapters only option.

Revision 1.22  2009/04/14 21:37:39  as-sgt
add const to HPI_MessageUDP implementation to match declaration

Revision 1.21  2009/04/07 15:35:55  as-tfe
Don't unload UDP DLL on subsys close.

Revision 1.20  2008/08/20 13:25:56  as-age
warnings

Revision 1.19  2008/07/15 01:15:35  as-ewb
merge from branc30711 - standalone hpi.h, hpi header includes use double quotes, remove win98

Revision 1.18.4.1  2008/07/04 05:31:20  as-ewb
switch to hpi_internal. Remove win16, dspcode array loader. hpi.h now standalone

Revision 1.18  2008/03/13 18:17:22  as-tfe
Remove TEXT() from HPI_DEBUG_LOG() format parameters.  It's now inside
the macro.

Revision 1.17  2008/03/12 23:32:53  as-ewb
Add level parameter to HPI_DEBUG_MESSAGE

Revision 1.16  2008/01/21 19:38:45  as-age
Support DLL unload on close.

Revision 1.15  2008/01/08 21:34:23  as-age
fix condition when asihpiudp32.dll is missing

Revision 1.14  2008/01/07 20:17:10  as-age
Correct import size.

Revision 1.13  2008/01/07 19:58:55  as-age
Use define for HPI_ETHERNET_TIMEOUT_MS.
Use HPI_MessageUDP() that passes a timeout parameter.

Revision 1.12  2007/12/03 19:24:25  as-age
Use DEBUG macros to output all strings.

Revision 1.11  2007/11/23 04:26:36  as-ewb
Ran Hpindent over all HPI files.

Revision 1.10  2007/11/05 16:47:18  as-age
Clarify network interface binding API.

Revision 1.9  2007/10/26 20:58:28  as-age
Add HPI_SubSysSetIPAddress() to set host IP address to use. This module does not
dynamically load asihpiudp32.dll UNTIL HPI_SubSysSetIPAddress() is called.

Revision 1.8  2007/10/18 20:37:26  as-tfe
Added compile time option to dynamically load and link to udp dll.  Define
HPI_BUILD_USEUDPDLL to use this option.
If this is not set and HPI_BUILD_MULTIINTERFACE
is defined then HPIUDP.C must be statically linked.

Revision 1.7  2007/08/14 15:13:05  as-age
Add msg debug output.

Revision 1.6  2007/06/19 14:07:18  as-age
Change hpimux.c to use hpidebug.c messages.

Revision 1.5  2007/06/11 20:10:57  as-age
Lots of changes. Mainly cheking in at this point so as to have a "way point". Still
needs more cleaning up.

Revision 1.4  2007/06/06 15:54:26  as-age
Simplify. This version works with ASX at least.

*/
