/******************************************************************************
Copyright (C) 1997-2010 AudioScience, Inc. All rights reserved.

This software is provided 'as-is', without any express or implied warranty.
In no event will AudioScience Inc. be held liable for any damages arising
from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must not
   claim that you wrote the original software. If you use this software
   in a product, an acknowledgment in the product documentation would be
   appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be
   misrepresented as being the original software.
3. This copyright notice and list of conditions may not be altered or removed
   from any source distribution.

AudioScience, Inc. <support@audioscience.com>

( This license is GPL compatible see http://www.gnu.org/licenses/license-list.html#GPLCompatibleLicenses )

  Firmware downloader for ASI2xxx network products

 (C) Copyright AudioScience Inc. 2005-2009
*******************************************************************************/
#include <string.h>
#include <stdlib.h>


#include "hpi.h"
#include "hpinet.h"
#include "hpifirmware.h"
#include <getopt.h>

#ifdef _MSC_VER
     #define snprintf _snprintf
#endif

#define MAX_FILES 4
#define MAC_STRING_SIZE 12

#ifndef MAX_PATH
#define MAX_PATH 260
#endif

struct fw_info {
	enum eHpiFirmwareId type;
	char path[MAX_PATH];
	char name[12];
};

static struct fw_info fw[MAX_FILES] = {
	{0,"","INVALID"},
	{HPI_FW_UPDATE, "", "Update"},
	{HPI_FW_FACTORY, "", "Factory"},
	{HPI_FW_BOOTLOADER, "", "Bootloader"}
};

/* Option variables */
static int num_files = 0;
static int shown_help = 0;
static int list = 0;
static int query = 0;
static int family = 0;
static int adapter_specified = 0;
static uint16_t wAdapterIndex = 0;
static uint16_t wAdapterFamily = 0;
static char szAdapterName[64];

static struct option long_options[] = {
	{"list", no_argument, 0, 'l'},
	{"adapter", required_argument, 0, 'a'},
	{"family", required_argument, 0, 'F'},
	{"factory", required_argument, 0, 'f'},
	{"update", required_argument, 0, 'u'},
	{"bootloader", required_argument, 0, 'b'},
	{"all", required_argument, 0, 'A'},
	{"query", no_argument, 0, 'q'},
	{"help", no_argument, 0, 'h'},
	{0, 0, 0, 0}
};

static const char *short_options = "la:F:f:u:b:A:qh?";

static const char *option_help[] = {
	"List adapters found on the network",
	"[index] Adapter index",
	"[Adapter .bin file family]",
	"[factory image filename] Download factory firmware (Danger!)",
	"[update image filename] Download update firmware",
	"[bootloader image filename ] Download bootloader firmware (Danger!)",
	"[factory and update image filename] Download BOOT2400.BIN, and both images with given file",
	"Query current hardware/firmware versions",
	"Show this text.\n\n"
	"For example:\n"
	" firmdown -l, will find the adapters on the network, listing them by <index>\n"
	" firmdown -a<index> -u firmware.bin, downloads firmware.bin into the spec'd adapter\n"
	" firmdown -a<index> -F2501 -udsp2501.bin, load dsp2501 to an ASI2416\n"
	" firmdown -a<index> -F2502 -udsp2502.bin, load dsp2502 to an ASI2501\n"
	" firmdown -a<index> -udsp2501.bin, load dsp2501 to an ASI2501\n"
};

static void help(char *progname)
{
	int i = 0;
	printf("\nUsage - %s [options]\n",
		progname);
	while (long_options[i].name != 0) {
		printf("--%s -%c %s\n",
			long_options[i].name,
			(char)(long_options[i].val), option_help[i]);
		i++;
	}
	printf("\n");
}

static void parse_options(int argc, char *argv[])
{
	int c;
    /*********** Parse the command line options ***************/
	while (1) {
		//int this_option_optind = optind ? optind : 1;
		int option_index = 0;

		c = getopt_long(argc, argv, short_options,
			long_options, &option_index);
		if (c == -1)
			break;

		switch (c) {
		case 0:
			printf("option %s", long_options[option_index].name);
			if (optarg)
				printf(" with arg %s", optarg);
			printf("\n");
			break;
		case 'a':
			wAdapterIndex = atoi(optarg);
			adapter_specified = 1;
			break;
		case 'F':
			family = 1;
			sscanf(optarg, "%hx", &wAdapterFamily);
			break;
		case 'u':
			strcpy(fw[HPI_FW_UPDATE].path, optarg);
			break;
		case 'f':
			strcpy(fw[HPI_FW_FACTORY].path, optarg);
			break;
		case 'b':
			strcpy(fw[HPI_FW_BOOTLOADER].path, optarg);
			break;
		case 'A':
			// program all (b,u,f) in one go
			strcpy(fw[HPI_FW_BOOTLOADER].path, "BOOT2400.BIN");	//assume this file is present in local dir
			strcpy(fw[HPI_FW_FACTORY].path, optarg);
			strcpy(fw[HPI_FW_UPDATE].path, optarg);
			printf("Programming with:\n b=%s\n f=%s\n u=%s\n",
				fw[HPI_FW_BOOTLOADER].path,
				fw[HPI_FW_FACTORY].path,
				fw[HPI_FW_UPDATE].path);
			break;
		case 'q':
			query = 1;
			break;
		case 'l':
			list = 1;
			break;
		case '?':
		case 'h':
			shown_help = 1;
			help(argv[0]);
			exit(0);
			break;

		default:
			printf("?? getopt returned character code 0%o ??\n", c);
		}
	}

	if (optind < argc) {
		// printf ("non-option ARGV-elements: ");
		while ((optind < argc) && (num_files < MAX_FILES)) {
			strncpy(fw[num_files].path, argv[optind++],
				MAX_PATH - 1);
			printf("%s image is %s\n", fw[num_files].name,
				fw[num_files].path);
			num_files++;
		}
	}
}

/*=================================================================*/
/**

*/
static void HPI_PrintError(
	uint16_t wError,
	uint16_t wSpecificError
)
{
	char errbuf[256];

	if (wError == 0) {

	} else {
		HPI_GetErrorText(wError, errbuf);
		printf(" Error%s, SE=%d\n", errbuf, wSpecificError);
	}
}

/*=================================================================*/
/**
Progress callback
*/

static void HPI_CALLBACK pc(char *state, u_int percent)
{
	if (percent)
		printf("%s %d%%", state, percent);
	else
		printf("%s", state);
}
/*=================================================================*/
/**
Map adapter names
*/
static const char *MapAdapterName(const int nAdapterType)
{
	snprintf(szAdapterName,sizeof(szAdapterName),"ASI%x, ",nAdapterType);
	return szAdapterName;
}


/*=================================================================*/
/**

*/
int main(int argc,	char *argv[])
{
	hpi_hsubsys_t *hSubSys;
	int i = 0;
	short error = 0;
	enum eHpiFirmwareId sel;
	int wNumAdapters;
	uint16_t wAdapterType=0;

	printf("Network Firmware Updater " HPI_VER_STRING " (c) AudioScience, Inc. 2009\n\n");
	parse_options(argc, argv);

	if (!shown_help)
		printf("To see full option list, use -h option\n\n");

	hSubSys = HPI_SubSysCreate();
	if (!hSubSys) {
		printf("ERROR: SubSysOpen Failed\n");
		return 1;
	}

	if (list)
	{

		error = HPI_SubSysGetNumAdapters(hSubSys, &wNumAdapters);

		if (!error && (wNumAdapters>0) )
		{
			printf("Choose one of these adapters using -a<index>:\n\n");
			for (i = 0; i < wNumAdapters; i++)
			{
				uint32_t dwAdapterIndex=0;
				uint32_t dwSerialNumber=0;
				uint16_t wAdapterType=0;
				uint16_t wNumOutStreams=0;
				uint16_t wNumInStreams=0;
				uint16_t wVersion=0;
				uint16_t wIPmsb=0;
				uint16_t wIPlsb=0;
				uint16_t wParam1=0;
				uint16_t nDspSwVer=0;
				uint16_t wDspSwVerRelease=0;
				char szIp[64];
				char cHwVerMajor='x';
				char cHwVerMinor='x';

				error = HPI_SubSysGetAdapter(hSubSys, i,
					&dwAdapterIndex, &wAdapterType);

				if(error) {
					printf("ERROR, adapter index %d, error=%d, on HPI_SubSysGetAdapter()\n",
						dwAdapterIndex,error);
					continue;
				}

				error = HPI_AdapterOpen(hSubSys, (uint16_t)dwAdapterIndex);
				if(error) {
					printf("ERROR, adapter index %d, error=%d, on HPI_AdapterOpen()\n",
						dwAdapterIndex,error);
					continue;
				}

				error = HPI_AdapterGetInfo(
					     hSubSys,
					     (uint16_t)dwAdapterIndex,
					     &wNumOutStreams,
					     &wNumInStreams,
					     &wVersion,
					     &dwSerialNumber,
					     &wAdapterType);

				if(error) {
					printf("ERROR, adapter index %d, error=%d, on HPI_AdapterGetInfo()\n",
						dwAdapterIndex,error);
					goto adapter_list_error;
				}
				cHwVerMajor = ((wVersion>>3)&0xF) + 'A';
				cHwVerMinor = wVersion&0x7;
				nDspSwVer = ((wVersion>>13)*100)+((wVersion>>7)&0x3F);

				error = HPI_AdapterGetProperty(
						hSubSys,
						(uint16_t)dwAdapterIndex,
						HPI_ADAPTER_PROPERTY_SOFTWARE_VERSION,
						&wParam1,
						&wDspSwVerRelease
						);
				if(error) {
					printf("ERROR, adapter index %d, error=%d, on HPI_AdapterGetProperty(SW_VER)\n",
						dwAdapterIndex,error);
					goto adapter_list_error;
				}


				error = HPI_AdapterGetProperty(hSubSys,
						(uint16_t)dwAdapterIndex,
						HPI_ADAPTER_PROPERTY_IP_ADDRESS,
						&wIPmsb,
						&wIPlsb );
				if(error) {
					printf("ERROR, adapter index %d, error=%d, on HPI_AdapterGetProperty(IP)\n",
						dwAdapterIndex,error);
					goto adapter_list_error;
				}

				sprintf(szIp,"%u.%u.%u.%u",
						(wIPmsb>>8) & 0xff,
						wIPmsb & 0xff,
						(wIPlsb>>8) & 0xff,
						wIPlsb & 0xff
						);

				if (!error)
					printf("%s Index=%d, "
						"S/N=%5d, "
						"Ver=%01d.%02d.%02d, "
						"Rev=%c%d, "
						"IP=%s \n",
					MapAdapterName(wAdapterType),
					dwAdapterIndex,
					dwSerialNumber,
					nDspSwVer/100, nDspSwVer%100, wDspSwVerRelease,
					cHwVerMajor,cHwVerMinor,
					szIp
					);

adapter_list_error:
				error = HPI_AdapterClose(hSubSys, (uint16_t)dwAdapterIndex);
				if(error) {
					printf("ERROR, adapter index %d, error=%d, on HPI_AdapterClose()\n",
						dwAdapterIndex,error);
					continue;
				}
			}
			printf("\nWhere the following abbreviations are used:\n"
				"Ver = Firmware Version\n"
				"Rev = Hardware Revision\n");
		}
		else
			printf("ERROR: No adapters found on the network\n");
		exit(0);
	}

	if(!adapter_specified)
		exit(0);

	//========= adapter open ==========
	error = HPI_AdapterOpen(hSubSys, wAdapterIndex);
	HPI_PrintError(error, 0);
	if (error)
		exit(1);

	{	//========= adapter get info ==========
		uint16_t wNumOutStreams;
		uint16_t wNumInStreams;
		uint16_t wVersion;
		uint32_t dwSerialNumber;

		error = HPI_AdapterGetInfo(hSubSys, wAdapterIndex,
			&wNumOutStreams, &wNumInStreams,
			&wVersion, &dwSerialNumber, &wAdapterType);
		HPI_PrintError(error, 0);

		printf("Adapter Type %4X \nIndex %d \nS/N %d\nHardware version %c%d\n", wAdapterType, wAdapterIndex, dwSerialNumber, ((wVersion >> 3) & 0xf) + 'A',	// Hw version major
			wVersion & 0x7	// Hw version minor
			//,((wVersion >> 13) * 100) + ((wVersion >> 7) & 0x3f)	// DSP code version
			);
	}

	if(!family) {
		//========= set adapter family if not on cmd line ==========
		switch (HPI_ADAPTER_FAMILY_ASI(wAdapterType))
		{
			case HPI_ADAPTER_FAMILY_ASI(0x2500):
				wAdapterFamily = wAdapterType;
				break;
			default:
				wAdapterFamily = HPI_ADAPTER_FAMILY_ASI(wAdapterType);
				break;
		}
	}

	if (query)
	for (sel = HPI_FW_UPDATE; sel <= HPI_FW_BOOTLOADER; sel++) {
		unsigned int version, checksum, length, adapter;

		error = HPI_AdapterFirmwareGetInfo(wAdapterIndex, sel,
			&version, &checksum, &length);

		if (error) {
			printf("Error getting %s image info\n", fw[sel].name);
			continue;
		}

		printf("%s image, version %3d, checksum %08x, length %d\n",
			fw[sel].name, version, checksum, length);

		if (strlen(fw[sel].path)) {
			error = HPI_FileFirmwareGetInfo(fw[sel].path,
				wAdapterFamily,
				0, /* family mask, not used now */
				&adapter,
				&version,
				&checksum,
				&length);

			if (error)
				HPI_PrintError(error, 0);
			else
				printf("file %s, version %3d, checksum %08x, length %d\n",
					fw[sel-1].name,  version, checksum, length);
		}
	}

	//========= program selected images to flash ==========
	for (sel = HPI_FW_UPDATE; sel <= HPI_FW_BOOTLOADER; sel++) {
		if (strlen(fw[sel].path)) {
			error = HPI_AdapterFirmwareDownload(wAdapterIndex,
				fw[sel].type, fw[sel].path,  wAdapterFamily, 0xFF00, pc);
			HPI_PrintError(error, 0);
		}
	}

	if (strlen(fw[HPI_FW_UPDATE].path))
		error = HPI_AdapterRestart(wAdapterIndex);

	if (error) {
		exit(1);
	}
	exit(0);
}

